//===================================================================//
//                                                                   //
//  JWPce Copyright (C) Glenn Rosenthal, 2002                        //
//  All rights reserved.                                             //
//                                                                   //
//  This is an index utility, and is intended to be used to generate //
//  index files for use with JWPce.  This utility process dictionary //
//  files in EDICT format using EUC or UTF-8 encoding.  The EDICT    //
//  format was developed by James William Breen.                     //
//                                                                   //   
//===================================================================//

//===================================================================
//
//  This utility was designed to index EDICT format dictionaries in
//  ECU or UTF-8 format.  
//
//  If you compile the utility make sure you allocate enough stack 
//  space or you will have problems (20MB or more).
//

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

//===================================================================
//
//  Type definitions.
//
typedef unsigned short KANJI;
typedef unsigned char  byte;

//===================================================================
//
//  Macros
//

//--------------------------------
//
//  Specific characters.
//
#define KANA_SLASH      0x213f           
#define KANA_QUESTION   0x2129
#define KANJI_BAD       0xffff          // Knaji used for invalid kanji

//--------------------------------
//
//  JIS based caracter pages.
//
#define BASE_ASCII      0x0000
#define BASE_SPECIAL1   0x2100
#define BASE_SPECIAL2   0x2200           
#define BASE_JASCII     0x2300
#define BASE_HIRAGANA   0x2400
#define BASE_KATAKANA   0x2500
#define BASE_GREEK      0x2600
#define BASE_CYRILIC    0x2700
#define BASE_KANJI      0x3000

//--------------------------------
//
//  Character type tests.
//
#define ISBASE(c,b)     (b == (c & 0xff00))                                 // Test for is JIS character in a specific base.
#define ISKANJI(c)      (c >= BASE_KANJI)                                   // Is kanji
#define ISASCII(c)      (ISBASE(c,BASE_ASCII))                              // Is ascii (including extended ascii)
#define ISHIRAGANA(c)   (ISBASE(c,BASE_HIRAGANA))                           // Is Hiragana
#define ISKATAKANA(c)   (ISBASE(c,BASE_KATAKANA))                           // Is katakana
#define ISKANA(c)       (ISHIRAGANA(c) || ISKATAKANA(c))                    // Is kana.
#define ISKANAQ(c)      (ISKANA(c) || (c == KANA_QUESTION))
#define ISEND(c)        ((c == '[') || (c == ']') || (c == '(') || (c == ')') || (c == ' ') || (c == ',') || (c == '/') || (c == KANA_SLASH))

//--------------------------------
//
//  UTF-8 encouding values. 
//
#define UTF8_VALUE1     0x00        // Value for set bits for single byte UTF-8 Code.
#define UTF8_MASK1      0x80        // Mask (i.e. bits not set by the standard) 0xxxxxxx
#define UTF8_WRITE1     0xff80      // Mask of bits we cannot allow if we are going to write one byte code
#define UTF8_VALUE2     0xc0        // Two byte codes
#define UTF8_MASK2      0xe0        // 110xxxxx 10yyyyyy
#define UTF8_WRITE2     0xf800      // Mask of mits we cannot allow if we are going to write two byte code
#define UTF8_VALUE3     0xe0        // Three byte codes    
#define UTF8_MASK3      0xf0        // 1110xxxx 10yyyyyy 10zzzzzz
#define UTF8_VALUE4     0xf0        // Four byte values
#define UTF8_MASK4      0xf8        // 11110xxx ----    (These values are not supported by JWPce).    
#define UTF8_VALUEC     0x80        // Continueation byte (10xxxxxx).
#define UTF8_MASKC      0xc0

//===================================================================
//
//  Static tables
//
//  These are primarally conversion talbes for processing Unicode data.
//
#define NUMBER_KANJIUNICODE     ((int) (sizeof(kanji_unicode)/sizeof(KANJI)))

static KANJI kanji_unicode[] = {    // Main kanji table.  These are unicode values
  #include "jwp_ukan.dat"           //   associated with jwp_info type indexes.
};

#define NUMBER_MISCUNICODE      ((int) (sizeof(misc_unicode)/sizeof(KANJI)))

static KANJI misc_unicode[] = {     // Main symbol group, corresponds to JIS codes 0x2100
  #include "jwp_umis.dat"
};

static KANJI cp1250[128] = {        // Eastern Europe code page
  #include "jwp_cp1250.dat"
};

static KANJI cp1251[128] = {        // Cyrillic code page
  #include "jwp_cp1251.dat"
};

static KANJI cp1252[128] = {        // USA, West Europe code page
  #include "jwp_cp1252.dat"
};

static KANJI cp1253[128] = {        // Greek code page
  #include "jwp_cp1253.dat"
};

static KANJI cp1254[128] = {        // Turkish code page
  #include "jwp_cp1254.dat"
};

static KANJI cp1255[128] = {        // Hebrew code page
  #include "jwp_cp1255.dat"
};

static KANJI cp1256[128] = {        // Arabic code page
  #include "jwp_cp1256.dat"
};

static KANJI cp1257[128] = {        // Baltic code page
  #include "jwp_cp1257.dat"
};

static KANJI cp1258[128] = {        // Vietnamese code page
  #include "jwp_cp1258.dat"
};

#define NUMBER_EXTUNICODE       128

//===================================================================
//
//  Global varaibles.  These include the state variables.
//
//  Note about the all, any, short varaibles.  Yes we don need three to get all the good combinatisons.
//
//      ALL_KANA   takes all kana so it does not matter what else you set.
//      ANY_KANA   takes kana of length two from anywhere in the word.
//      SHORT_KANA takes kana words that are 1 character.
//
//      Setting ANY_KANA and SHORT_KANA is not the same thing as ALL_KANA.
//
//      Example: DENKIYA / HA
//
//          ALL_KANA        DE N KI YA / HA
//
//          ANY_KANA        DE N KI    / --
//
//          SHORT_KANA      --         / HA
//
//          ANY & SHORT     DE N KI    / HA
//
//          normal          DE
//

int    all_kana    = false;     // Include all kana, not just kana at the beginning of the word.
int    all_ascii   = false;     // Include all ascii, not just at the beginning of the word.
int    any_kana    = false;     // Include any kana that matches the length requirements, not just at the beginning of a word.
int    any_ascii   = false;     // Include any ascii that matches the length requirement, not just at the beginning of a word
int    skip_notes  = false;     // Skip anything in parentheses.
int    short_kana  = false;     // Include kana with less than 2 kana.
int    short_ascii = false;     // Include ascii with less than 3 characters.
int    ascii2      = false;     // Shortents the ascii character acceptance to 2 characters.
int    utf_format  = false;     // Input file is in UTF-8 format.
int    test        = false;     // Just scan the data do not sort or save.
int    warn        = true;      // Display warning messages.
KANJI *ext_unicode = cp1252;    // USA and Western Europe code page (UTF only)
byte  *dict        = NULL;      // Pointer to memory copy of the dictionary
byte **keys        = NULL;      // Pointer to the keys array.

//===================================================================
//
//  Static routines
//
int   compare     (byte *key1,byte *key2);          // Compare two keys
long  file_size   (char *name);                     // Get size of file in bytes.
KANJI get_char    (byte *&ptr);                     // Get character in current format and advance pointer.
KANJI get_test    (byte *&ptr);                     // Get character for comparison (checks for end of key)
void  quicksort   (long low,long high);             // Sort routine.
void  terminate   (char *message,char *param);      // Terminal error handler.
int   unicode2jis (int ch,int bad);                 // Convert UNICODE to JIS
int   utf2jis     (byte *&ptr);                     // Convert UTF-8 to JIS
void  warning     (char *message,byte *text);       // Display warning message.

//--------------------------------
//
//  This routine compiares two keys and indicates the difference.
//
//      key1,key2 -- Keys to be compared.
//
//      RETURN    -- Indicates the difference
//                      < 0 -- key1 < key2
//                      = 0 -- key1 = key2
//                      > 0 -- key1 > key2
//
int compare (byte *key1,byte *key2) {
  KANJI ch1,ch2;
  do {
    ch1 = get_test(key1);
    ch2 = get_test(key2);
  } while (ch1 && ch2 && (ch1 == ch2));
  return (ch2-ch1);
}

//--------------------------------
//
//  Get the size of a file.
//
//      name   -- Name of file to check.
//      
//      RETURN -- Size in bytes.
//
long file_size (char *name) {
  FILE *file;
  long  size;
  if (!(file = fopen(name,"rb"))) return (0);
  fseek  (file,0,SEEK_END);
  size = ftell(file);
  fclose (file);
  return (size);
}

//--------------------------------
//
//  This rotuine gets a character for testing in the comparison routine.
//  This is really a variaent of the get_char() routine that deals with the
//  end of the word.  If the end of the word is detected, a value of zero is
//  returned.
//
//      ptr    -- On entry this points to the buffer to be examened.  On exit,
//                this will point to the next character.  Note that characters
//                can vary in lenght from 1 to 4 bytes.
//      RETURN -- Character extracted or zero if this is the end of a word.
//
KANJI get_test (byte *&ptr) {
  KANJI ch;
  ch = get_char(ptr);
  if (ISASCII(ch)) ch = tolower(ch);
  if (ISKATAKANA(ch)) ch = BASE_HIRAGANA | (ch & 0xff);
  if (ISEND(ch)) return (0);
  return (ch);
}

//--------------------------------
//
//  Quicksort used for sotring the list of keys.
//
//      low,high -- Bounds for the sort.  Call with arguments (0,count).
//
//  This is a recursive routine and can use lots of stack space.
//
#define SWAP(x,y) { ptr = keys[x]; keys[x] = keys[y]; keys[y] = ptr; }

void quicksort (long low,long high) {
  long i,last;
  byte *ptr;
  if (low >= high) return;                          // Done.
  i    = (low+high)/2;                              // This is not necessary, but is done in case the list is in order (Quicksort does not
  SWAP (low,i);                                     //   work well when the list is in order, but this should not happen here.
  last = low;
  for (i = low+1; i <= high; i++) {                 // Partition.  Move all elements before selected (low, but was mid) element before the 
    if (compare(keys[low],keys[i]) < 0) {           //   partition marker.
      last++;
      SWAP (i,last);
    }
  }
  SWAP      (low,last);                             // Swap partion marker to end, between lists, since this element is were it belongs.
  quicksort (low,last-1);                           // Sort both halves of the list.
  quicksort (last+1,high);
  return;
}

//--------------------------------
//
//  Abort routine.  Deallocate resource.  Put up a termiantion message
//  and shut down.  This is normally called after an error.
//
//      message -- Message to display (in printf format)./
//      param   -- Parameter for the printf message.
//
void terminate (char *message,char *param) {
  printf ("\n\n");
  printf (message,param);
  printf ("\n\n");
  if (dict) free (dict);
  if (keys) free (keys);
  exit (1);
}

//--------------------------------
//
//  This routine converts from Unicode to JIS codes.  This is one half
//  of the core of the Unicode support.
//
//      ch     -- Unicode character.
//      bad    -- Character returned for untraslateable UNICODE.
//
//      RETURN -- JIS character code.
//
int unicode2jis (int ch,int bad) {
  int i;
  if ((ch <= 0x007e)                  ) return (ch);                // ASCII
  if ((ch >= 0x3041) && (ch <= 0x3093)) return (ch-0x3041+0x2421);  // Hiragana
  if ((ch >= 0x30a1) && (ch <= 0x30f6)) return (ch-0x30a1+0x2521);  // Katakana
  if ((ch >= 0x0391) && (ch <= 0x03c9)) {                           // Greek
    if (ch <= 0x03a1) return (ch-0x0391+0x2621);
    if (ch == 0x03a2) return (0);
    if (ch <= 0x03a9) return (ch-0x0392+0x2621);
    if (ch <  0x03b1) return (0);
    if (ch <= 0x03c1) return (ch-0x0391+0x2621);
    if (ch == 0x03c2) return (0);
    if (ch <= 0x03c9) return (ch-0x0392+0x2621);
    return (0);
  }
  if (ch == 0x0401) return (0x2727);                                // Cyrillic
  if (ch == 0x0451) return (0x2757);
  if ((ch >= 0x0410) && (ch <= 0x044f)) {
    if (ch <= 0x0415) return (ch-0x0410+0x2721);
    if (ch <= 0x042f) return (ch-0x0416+0x2728);
    if (ch <= 0x0435) return (ch-0x0430+0x2751);
    if (ch <= 0x044f) return (ch-0x0436+0x2758);
  }
  for (i = 0; i < NUMBER_KANJIUNICODE; i++) {                       // The kanji
    if (kanji_unicode[i] == ch) {
      ch = i/94;
      return (((ch+0x30) << 8) | (i-ch*94)+0x21);
    }
  }
  for (i = 0; i < NUMBER_MISCUNICODE; i++) {                        // Symbol and pucntuation (main group)
    if (misc_unicode[i] == ch) {
      switch (i/94) {
        case 0: return (0x2121+i);
        case 1: return (0x2221+i-94);
        case 2: return (0x2821+i-94-94);
      }
    }
  }
  for (i = 0; i < NUMBER_EXTUNICODE; i++) {                         // Extneded ascii codes.
    if (ext_unicode[i] == ch) return (i+0x80);
  }
//
//  The Japanese ASCII needs to be here because some puncutation is stored within gaps
//   in the JASCII table, and if we move this before the misc codes they will not be found!    
//
  if ((ch >= 0xff10) && (ch <= 0xff5a)) return (ch-0xff10+0x2330);  // Japanese ASCII.
  if ((ch == 0xfeff) || (ch == 0xfffe)) return (ch);                // UNICODE file markers.
  return (bad);                                                     // Invalid or untranslated UNICODE
}                               

//--------------------------------
//
//  Convert UTF-8 data into JIS value.  Because UTF-8 can take up to 4 bytes to store a value, 
//  this is a varing length conversion.  Further, note that 4 byte values are not supported, since
//  they lead to extended JIS which is not supported.
//
//  This routine is called with a pointer to a data array.  A single JIS value is extracted,
//  and the pointer is advanced to the next data to be analyzed.
//
//      ptr    -- On entry a pointer to the start of a UTF-8 encoded character.  On exit, this
//                will point to the next UTF character.
//
//      RETURN -- JIS charactger value or KANJI_BAD to indicate an error.
//
int utf2jis (byte *&ptr) {
  int i;
  if      (UTF8_VALUE1 == (*ptr & UTF8_MASK1)) {
    i = *ptr++;
  }
  else if (UTF8_VALUE2 == (*ptr & UTF8_MASK2)) {
    i = ((ptr[0] & 0x1f) << 6) | (ptr[1] & 0x3f);
    ptr += 2;
  }
  else if (UTF8_VALUE3 == (*ptr & UTF8_MASK3)) {
    i = ((ptr[0] & 0xf) << 12) | ((ptr[1] & 0x3f) << 6) | (ptr[2] & 0x3f);
    ptr += 3;
  }
  else if (UTF8_VALUE4 == (*ptr & UTF8_MASK4)) {
    i = 0xffff;
    ptr += 4;
  }
  return (unicode2jis(i,KANJI_BAD));
}

//--------------------------------
//
//  Utility routine to display a warning message.  This routine takes care of
//  of dealing with the line spacing and some other niceties.
//
//      message -- printf() style message.
//      text    -- Parameter for message.
//
void warning (char *message,byte *text) {
  static int first = true;
  int  i;
  char buffer[512];
  if (!warn) return;
  if (first) printf ("\n");
  first = false;
  for (i = 0; (i < sizeof(buffer)-1) && (text[i] != '\n'); i++) buffer[i] = text[i];
  buffer[i] = 0;  
  printf (message,buffer);
  return;
}






#define ISASCII1(c)  (ISASCII(c) && (isalpha(c) || (c > 0x0080)))
#define ISASCII2(c)  (ISASCII(c) && !ISEND(c))
#define ISDIGIT(c)   (ISASCII(c) && (isdigit(c)))
//#define ISASCII2(c)  (ISASCII(c) && (isalnum(c) || (c == '-') || (c == '.') || (c > 0x0080)))
#define KANA_LONGVOWEL  0x213c           
#define ISKANA1(c)  (ISKANA(c) || (c == KANA_QUESTION))
#define ISKANA2(c)  (ISKANA(c) || (c == KANA_QUESTION) || (c == KANA_LONGVOWEL))           



#define DEBUG   20000
#define DEBUG_NAME  "DEBUG1.TXT"





//--------------------------------
//
//  This routine gets a character from some location in the dictionary.  The
//  pointer is advanced to the next character.  This routine handles EUC and 
//  UTF-8.  Actually, this is the only location where there is a difference 
//  between the formats.
//
//      ptr    -- On entry this points to the lcoation to extract a character.
//                On exit this points to the next character.  The length of a
//                character can be 1 to 4 bytes.
//
//      RETURN -- Extracted character in JIS.
//
KANJI get_char (byte *&ptr) {
  KANJI ch;
  if (utf_format) return (utf2jis(ptr));
  if (!(*ptr & 0x80)) return (*ptr++);
  ch = *ptr++;
  return (((ch << 8) | *ptr++) & 0x7f7f);
}

//--------------------------------
//
//  The actual program.
//
#define YESNO(x) ((x) ? "YES" : "NO")  

int main (int argc,char *argv[]) {
  int i;
  int  count;  
  char *name;
  FILE *file;
//
//  Initialization code.  Check parameters, and handle simple errors.
//  
  int error = false;
  printf ("JINDEX Utility Verison 1.0 copyright (C) 2002 glenn rosenthal.\n\n");
  for (i = 2; i < argc; i++) {
    if      (!stricmp(argv[i],"allkana"   )) all_kana    = true;
    else if (!stricmp(argv[i],"anykana"   )) any_kana    = true;
    else if (!stricmp(argv[i],"shortkana" )) short_kana  = true;
    else if (!stricmp(argv[i],"allascii"  )) all_ascii   = true;
    else if (!stricmp(argv[i],"anyascii"  )) any_ascii   = true;
    else if (!stricmp(argv[i],"shortascii")) short_ascii = true;
    else if (!stricmp(argv[i],"skipnotes" )) skip_notes  = true;
    else if (!stricmp(argv[i],"utf"       )) utf_format  = true;
    else if (!stricmp(argv[i],"test"      )) test        = true;
    else if (!stricmp(argv[i],"nowarn"    )) warn        = false;
    else if (!stricmp(argv[i],"1250"      )) ext_unicode = cp1250;
    else if (!stricmp(argv[i],"1251"      )) ext_unicode = cp1251;
    else if (!stricmp(argv[i],"1253"      )) ext_unicode = cp1253;
    else if (!stricmp(argv[i],"1254"      )) ext_unicode = cp1254;
    else if (!stricmp(argv[i],"1255"      )) ext_unicode = cp1255;
    else if (!stricmp(argv[i],"1256"      )) ext_unicode = cp1256;
    else if (!stricmp(argv[i],"1257"      )) ext_unicode = cp1257;
    else if (!stricmp(argv[i],"1258"      )) ext_unicode = cp1258;
    else                                     error       = true;
  }
//
//  Process errors
//  
  if (error || (argc < 2)) {
    printf ("USSAGE:  JINDEX <file> [ <flags> ]\n");
    printf ("  ALLKANA     -- Include all kana (not just at beggining of a word).\n");
    printf ("  ANYKANA     -- Include all kana of length 2 (not just beginning of word).\n");
    printf ("  SHORTKANA   -- Include short kana words (only full words of 1 kana).\n");
    printf ("  ALLASCII    -- Include all ASCII (not just at the beginning of a word).\n");
    printf ("  ANYASCII    -- Include all ASCII of lneght 3 (not just at the beginning of the word).\n");
    printf ("  SHAORTASCII -- Include short ASCII words (only full words of 1 or 2 ASCII).\n");
    printf ("  ASCII2      -- Set acceptance length for ASCII to 2 characters (not 3).\n");
    printf ("  SKIPNOTES   -- Skip entries in parethese.\n");
    printf ("  UTF         -- Input file is in UTF-8 format.\n");
    printf ("  TEST        -- Scan and process file, but don't save (used to check index size).\n");
    printf ("  NOWARN      -- Suppress warning messages.\n");
    printf ("  1250        -- Eastern Europe code page (UTF-8 only).\n");
    printf ("  1251        -- Cyrillic code page (UTF-8 only).\n");
    printf ("  1253        -- Greek code page (UTF-8 only).\n");
    printf ("  1254        -- Turkish code page (UTF-8 only).\n");
    printf ("  1255        -- Hebrew code page (UTF-8 only).\n");
    printf ("  1256        -- Arabic code page (UTF-8 only).\n");
    printf ("  1257        -- Baltic code page (UTF-8 only).\n");
    printf ("  1258        -- Vietnamese code page (UTF-8 only).\n");
    printf ("\n");
    return (1);
  }
  printf ("all_kana=%s  any_kana=%s  short_kana=%s\n",YESNO(all_kana),YESNO(any_kana),YESNO(short_kana));
  printf ("all_ascii=%s  any_ascii=%s  short_ascii=%s  ascii2=%s  skip_notes=%s\n",YESNO(all_ascii),YESNO(any_ascii),YESNO(short_ascii),YESNO(ascii2),YESNO(skip_notes));
  printf ("encode=%s  warnings=%s  test=%s  code_page=",utf_format ? "UTF-8" : "EUC",YESNO(warn),YESNO(test));
  if      (ext_unicode == cp1250) printf ("1250 (Eastern Europe)\n");
  else if (ext_unicode == cp1251) printf ("1251 (Cyrillic)\n");
  else if (ext_unicode == cp1252) printf ("1252 (USA and Western Europe)\n");
  else if (ext_unicode == cp1253) printf ("1253 (Greek)\n");
  else if (ext_unicode == cp1254) printf ("1254 (Turkish)\n");
  else if (ext_unicode == cp1255) printf ("1255 (Hebrew)\n");
  else if (ext_unicode == cp1256) printf ("1256 (Arabic)\n");
  else if (ext_unicode == cp1257) printf ("1257 (Baltic)\n");
  else if (ext_unicode == cp1258) printf ("1258 (Vietnamese)\n");
  else                            printf ("UNKNOWN!\n");
  printf ("\n");
//
//  Read in file and allocate arrays.
//
  name = argv[1];  
  printf ("Reading input file %s ... ",name);
  if (!(i = file_size(argv[1]))) terminate ("\n\nERROR -- Unable to open file %s",name);
  dict  = (byte *)  calloc(i+10,1);
  keys  = (byte **) calloc(i+10,sizeof(long));
  count = 0;
  if (!dict || !keys) terminate ("ERROR -- Unable to allocate memory",NULL);
  if (!(file = fopen(argv[1],"rb"))) terminate ("ERROR -- Unable to open file %s",name);
  if (i != fread (dict,1,i,file)) { fclose (file); terminate ("ERROR -- Error reading file %s",name); }
  fclose (file);
  printf ("done\n");
  keys[count++] = dict+i-1;         // This adds a blank entry to the list.  This will be sorted to the front.
//
//  Select the keys.
//
  byte *ptr,*ptr2,*ptr3;
  KANJI ch,ch2;
  int in_note       = false;        // Indicates in parenthetical note.
  int found_meaning = false;        // Indicates we have found the meaning on this line.
  int in_ascii      = false;        // In ascii word.
  int in_jword      = false;        // In Japanese word.
  byte *meaning = (byte *) "";
  printf ("Creating key entires ... ");
  for (ptr = dict; *ptr; ptr = ptr2) {
    ptr2 = ptr;
    ch   = get_char(ptr2);
    switch (ch) {
      case '(':
           in_ascii      = false;
           in_jword      = false;
           in_note       = true;
           break;
      case ')':
           in_ascii      = false;
           in_jword      = false;
           in_note       = false;
           break;
      case '\n':
      case '\r':
           if (!found_meaning) warning ("WARNING -- syntatical error, no meaning found [%s]\n",meaning);
           if (in_note       ) warning ("WARNING -- Umatched parentheses [%s]\n",meaning);
           in_note       = false;
           found_meaning = false;
           in_ascii      = false;
           in_jword      = false;
           break;
      case '/':
           if (!found_meaning) { found_meaning = true; meaning = ptr; }
           in_ascii      = false;
           in_jword      = false;
           break;
      case ',':                                                                             // , so check for word end.
           ptr3 = ptr2;                             
           ch   = get_char(ptr3);
           if (in_ascii && ISDIGIT(ch)) break;                                              // In digit so this something like 100,000.  Don't abort.
           in_ascii      = false;
           in_jword      = false;
           break;
      case KANA_SLASH:                                                                      // Always end word.
      case ' ':
      case '[':
      case ']':
           in_ascii      = false;
           in_jword      = false;
           break;
      default:
           if (in_note && skip_notes) break;                                                // In note so skip this.
           if (ISASCII1(ch)) {                                                              // ASCII
             ptr3 = ptr2;
             ch   = get_char(ptr3);
             ch2  = get_char(ptr3);
             if      (all_ascii) keys[count++] = ptr;                                       //   Take all ASCII
             else if (in_ascii && !any_ascii) { }                                           //   In a word, but don't take all so skip
             else if (ISDIGIT(ch)) { in_ascii = true; keys[count++] = ptr; }                //   Take all number beginning words.
             else if (ISASCII2(ch) && ISASCII2(ch2)) { in_ascii = true; keys[count++] = ptr; }  //   Three ascii in a row so mark in word and index.
             else if (ascii2 && ISASCII2(ch)) { in_ascii = true; keys[count++] = ptr; }     //   Set to accept 2 character words.
             else if (!in_ascii && short_ascii && ISEND(ch)) keys[count++] = ptr;           //   Short 1 character word
             else if (!in_ascii && short_ascii && ISASCII2(ch) && ISEND(ch2)) { in_ascii = true; keys[count++] = ptr; }     // Short 2 character word.
             else    { }                                                                    //   Ascii followed by something else so skip.
           }
           else if (ISKANA1(ch)) {                                                          // Kana
             ptr3 = ptr2;
             ch   = get_char(ptr3);
             if      (all_kana) keys[count++] = ptr;                                        //   Take all kana
             else if (in_jword && !any_kana) { }                                            //   In a word, but don't take all so skip
             else if (ISKANA2(ch)) { in_jword = true; keys[count++] = ptr; }                //   Two kana in a row, so mark in word and index.
             else if (!in_jword && short_kana && ISEND(ch)) keys[count++] = ptr;            //   Short kana word
             else    { }                                                                    //   Kana followed by something else so skip.
           }
           else if (ISBASE(ch,BASE_SPECIAL1) || ISBASE(ch,BASE_SPECIAL2)) {                 // Special characters are take if they are a single character
             ptr3 = ptr2;
             ch   = get_char(ptr3);
             if (ISEND(ch)) keys[count++] = ptr;
           }
           else if (ch == KANJI_BAD) { }                                                    // Skip any kanji we could not decode.
           else if (ISKANJI(ch)) {                                                          // Kanji and special characters.
             keys[count++] = ptr;
             in_jword      = true;
           }
           else if (ISBASE(ch,BASE_JASCII )) { }
           else if (ISBASE(ch,BASE_GREEK  )) { }
           else if (ISBASE(ch,BASE_CYRILIC)) { }
           else { }                                                                         // Should be no characters here, but you never know
           break;
    }
  }
  printf ("[%d] done\n",count);
//
//  Test processing
//  
  if (test) {
    printf ("\nIndex file size %d\n\n",count*sizeof(long));
    return (0);
  }
//
//
//  
  printf    ("Sorting database (takes time!) ... ");
  quicksort (0,count-1);
  printf    ("done\n");
//
//  Test code to read in an index.  This then can be processed through the debug. 
//
//  I originally wrote this to check the exact format of EDICT index file, but decided to keep it in case it was needed.
//
#if 0
  i     = file_size ("edit.jdx");
  file  = fopen("edict.jdx","rb");
  count = i/4;
  fread  (keys,i,1,file);
  fclose (file);
  for (i = 0; i < count; i++) keys[i] = dict+((long *) keys)[i]-1;
#endif
//
//  Debuging.
//
//  Generates an EUC file containning the location of each index and then to the end of the line.
//
//  This file is only generated if DEBUG is defined.  Further, the value of DEBUG defines the 
//  number of elements written into the file.  This can be set to "count" to generate a complete
//  file.  This can be a really large file, however.
//
#ifdef DEBUG
  printf ("Creating debug file [%s] ... ",DEBUG_NAME);
  if (!(file = fopen(DEBUG_NAME,"wb"))) terminate ("ERROR -- Unable to open debug file %s",DEBUG_NAME);
  for (i = 0; i < DEBUG; i++) {
    for (ptr = keys[i]; *ptr && (*ptr != '\n'); ptr++) fputc(*ptr,file);
    fputc ('\n',file);
  }
  fclose (file);
  printf ("done\n");
#endif DEBUG  
//
//  Write index file
//
  char buffer[512];
  long offset;
  strcpy (buffer,name);
  for (i = (int) strlen(buffer)-1; (i >= 0) && (buffer[i] != '\\') && (buffer[i] != '.'); i--);
  if (buffer[i] == '.') buffer[i] = 0;
  strcat (buffer,".jdx");
  printf ("Writing index file [%s] ... ",buffer);
  if (!(file = fopen(buffer,"wb"))) terminate ("ERROR -- Unable to open index file %s",buffer);
  for (i = 0; i < count; i++) {
    offset = (long) (keys[i]-dict);
    offset++;                                               // This is in the EDICT standard.
    if (1 != fwrite(&offset,sizeof(long),1,file)) terminate ("Error -- Error writing to index file %s",buffer);
  }
  fclose (file);
  printf (" [%ld] done\n\n",count*sizeof(long));
  return (0);
}

