//===================================================================//
//                                                                   //
//  KINFO Copyright (C) Glenn Rosenthal, 1998-2002,2003.             //
//  All rights reserved.                                             //
//                                                                   //   
//  The KANJIDIC file this program extracts information from was     //
//  compiled by Jim Breen, who holds the copyright on it.            //
//                                                                   //
//===================================================================//

//===================================================================
//
//  This program was developed to extract information from Jim Breen's 
//  KANJIDIC and convert this inforamtion into a from that JWPce could
//  use.  
//
//  The program takes no imput arguments at the moment, and produces 
//  several output fils.  The program must be run from the same location
//  as KANJIDIC so it can find the kanji dictionary database.
//
//  The oputput files are as follows:
//
//      KANJI_FREQ.EUC -- This is the kanji listed by the usage 
//                        fequencies compiled by Jack Halpern.
//                        This is a EUC file listing the kanji by 
//                        the most fequent to the least frequent.
//                        The kanji are listed in blocks of 100, with
//                        numerical indicators.  See the copyright
//                        notices concerting the use of this file.
//
//      KANJINFO.DAT   -- Kanji information file with nanori, meanings
//                        kuy-yomi, on-yomi, and PinYin.
//      KANJINFO.MED   -- Kanji information file with meanings, kun-yomi,
//                        on-yomi, and PinYin.
//      KANJINFO.SML   -- Kanji information file with meanings, kun-yomi
//                        and on-yomi.   
//
//      JWP_UNIC.DAT   -- Kanji to Unicode conversion table.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include "kinfo.h"

//===================================================================
//
//  Static data.
//
#define NUMBER_BIGKANJI 7000


static int count = 0;               // Number of kanji read.
static ulong offset;

kinfo_big kanji[NUMBER_BIGKANJI];   // Information for each kanji.

//===================================================================
//
//  Encoding conversion routines.  Most of these are taken directly from JWPce.
//
static int translate   (char *buffer,int encode);       // Translite KANJIDIC line from input encoding to JIS mixed
static int unicode2jis (int ch);                        // Convert UNICODE to JIS
static int utf2unicode (char *&ptr);                    // Convert UTF-8 to UNICODE

#define NUMBER_KANJIUNICODE     ((int) (sizeof(kanji_unicode)/sizeof(KANJI)))

static KANJI kanji_unicode[] = {    // Main kanji table.  These are unicode values
  #include "jwp_ukan.dat"           //   associated with jwp_info type indexes.
};

#define NUMBER_MISCUNICODE      ((int) (sizeof(misc_unicode)/sizeof(KANJI)))

static KANJI misc_unicode[] = {     // Main symbol group, corresponds to JIS codes 0x2100
  #include "jwp_umis.dat"
};

//--------------------------------
//
//  Converts a lien from whatever input encoding to a mixed format encoding.  The temporary encoding has
//  all text converted to EUC except the meanings fields which are still in UTF-8.  If the input data is 
//  EUC then the meanins can only contain ASCII so UTF-8 and EUC are identica.
//
//      buffer -- Buffer to translate.  The return will be placed in this buffer.
//      encode -- Encoding system to be used.  (ENCODE_EUC or ENCODE_UTF8).
//
//      RETURN -- Nonzero return is a an error, usually an invalid cahracter.
//
static int translate (char *buffer,int encode) {
  int   i,meaning = false;                          // Set to true when proceesing a meaning.
  char *in,*out,temp[5000];
  if (encode == ENCODE_EUC) return (false);         // Already EUC so we have nothing to do.
  strcpy (temp,buffer);                             // Copy buffer to temp space.
  for (out = buffer, in = temp; *in; ) {
    switch (*in) {                                  // Check for meanings
      case '#': meaning = true ; break;             // This is really a comment.  May have to change this latter.
      case '{': meaning = true ; break;
      case '}': meaning = false; break;
      default:                   break;
    }
    if (meaning) *out++ = *in++;
      else {
        i = unicode2jis(utf2unicode(in));
        if (!i) return (true);                      // Invalid character found.
        if ((i == 0xffef) || (i == 0xfeff)) continue;
        if (i < 0x80) *out++ = i;
          else {
            *out++ = ((i & 0xff00) >> 8) | 0x80;
            *out++ =  (i & 0x00ff)       | 0x80;
          }
      }
  }
  *out++ = 0;
  return (false);
}

//--------------------------------
//
//  This routine converts from Unicode to JIS codes.  This is one half
//  of the core of the Unicode support.
//
//      ch     -- Unicode character.
//
//      RETURN -- JIS character code.
//
static int unicode2jis (int ch) {
  int i;
  if ((ch <= 0x007e)                  ) return (ch);                // ASCII
  if ((ch >= 0x3041) && (ch <= 0x3093)) return (ch-0x3041+0x2421);  // Hiragana
  if ((ch >= 0x30a1) && (ch <= 0x30f6)) return (ch-0x30a1+0x2521);  // Katakana
  if ((ch >= 0x0391) && (ch <= 0x03c9)) {                           // Greek
    if (ch <= 0x03a1) return (ch-0x0391+0x2621);
    if (ch == 0x03a2) return (0);
    if (ch <= 0x03a9) return (ch-0x0392+0x2621);
    if (ch <  0x03b1) return (0);
    if (ch <= 0x03c1) return (ch-0x0391+0x2621);
    if (ch == 0x03c2) return (0);
    if (ch <= 0x03c9) return (ch-0x0392+0x2621);
    return (0);
  }
  if (ch == 0x0401) return (0x2727);                                // Cyrillic
  if (ch == 0x0451) return (0x2757);
  if ((ch >= 0x0410) && (ch <= 0x044f)) {
    if (ch <= 0x0415) return (ch-0x0410+0x2721);
    if (ch <= 0x042f) return (ch-0x0416+0x2728);
    if (ch <= 0x0435) return (ch-0x0430+0x2751);
    if (ch <= 0x044f) return (ch-0x0436+0x2758);
  }
  for (i = 0; i < NUMBER_KANJIUNICODE; i++) {                       // The kanji
    if (kanji_unicode[i] == ch) {
      ch = i/94;
      return (((ch+0x30) << 8) | (i-ch*94)+0x21);
    }
  }
  for (i = 0; i < NUMBER_MISCUNICODE; i++) {                        // Symbol and pucntuation (main group)
    if (misc_unicode[i] == ch) {
      switch (i/94) {
        case 0: return (0x2121+i);
        case 1: return (0x2221+i-94);
        case 2: return (0x2821+i-94-94);
      }
    }
  }
#if 0
  for (i = 0; i < NUMBER_EXTUNICODE; i++) {                         // Extneded ascii codes.
    if (ext_unicode[i] == ch) return (i+0x80);                      // This section is disbled because extneded ASCII codes should not be here.
  }
#endif
//
//  The Japanese ASCII needs to be here because some puncutation is stored within gaps
//  in the JASCII table, and if we move this before the misc codes they will not be found!    
//
  if ((ch >= 0xff10) && (ch <= 0xff5a)) return (ch-0xff10+0x2330);  // Japanese ASCII.
//
//  Unicode begin and end markers
//
  if ((ch == 0xfeff) || (ch == 0xfffe)) return (ch);
//
//  Error return value.
//
  return (0);           
}                       

//--------------------------------
//
//  Convert UTF-8 data into UNICODE value.  Because UTF-8 can take up to 4 bytes to store a value, 
//  this is a varing length conversion.  Further, note that 4 byte values are not supported, since
//  they lead to extended UNICODE which is not supported.
//
//  This routine is called with a pointer to a data array.  A single UNICODE value is extracted,
//  and the pointer is advanced to the next data to be analyzed.
//
//      ptr    -- On entry a pointer to the start of a UTF-8 encoded character.  On exit, this
//                will point to the next UTF character.
//
//      RETURN -- UNICODE charactger value or 0xffff to indicate an error.
//
                                    // UTF-8 Flags and values
#define UTF8_VALUE1     0x00        // Value for set bits for single byte UTF-8 Code.
#define UTF8_MASK1      0x80        // Mask (i.e. bits not set by the standard) 0xxxxxxx
#define UTF8_WRITE1     0xff80      // Mask of bits we cannot allow if we are going to write one byte code
#define UTF8_VALUE2     0xc0        // Two byte codes
#define UTF8_MASK2      0xe0        // 110xxxxx 10yyyyyy
#define UTF8_WRITE2     0xf800      // Mask of mits we cannot allow if we are going to write two byte code
#define UTF8_VALUE3     0xe0        // Three byte codes    
#define UTF8_MASK3      0xf0        // 1110xxxx 10yyyyyy 10zzzzzz
#define UTF8_VALUE4     0xf0        // Four byte values
#define UTF8_MASK4      0xf8        // 11110xxx ----    (These values are not supported by JWPce).    

static int utf2unicode (char *&ptr) {
  int i;
  if      (UTF8_VALUE1 == (*ptr & UTF8_MASK1)) {
    i = *ptr++;
  }
  else if (UTF8_VALUE2 == (*ptr & UTF8_MASK2)) {
    i = ((ptr[0] & 0x1f) << 6) | (ptr[1] & 0x3f);
    ptr += 2;
  }
  else if (UTF8_VALUE3 == (*ptr & UTF8_MASK3)) {
    i = ((ptr[0] & 0xf) << 12) | ((ptr[1] & 0x3f) << 6) | (ptr[2] & 0x3f);
    ptr += 3;
  }
  else if (UTF8_VALUE4 == (*ptr & UTF8_MASK4)) {
    i = 0xffff;
    ptr += 4;
  }
  return (i);
}

//===================================================================
//
//  Basic parsing routines.
//
static void  add_byte       (byte *buffer,int &pos,int value);  // Add byte to extra stream
static void  add_code       (byte *buffer,int &pos,int value);  // Add code value to extra buffer
static void  add_entry      (byte *buffer,int &pos,int code,int value); // Add a code and short value (the most common entry).
static void  add_long       (byte *buffer,int &pos,long value); // Add long value to extra buffer
static void  add_short      (byte *buffer,int &pos,unsigned int value); // Add a short value to the extra buffer
static int   code_to_jis    (int code);                         // Convert JWPce compressed code to JIS code
static char *decode_pinyin  (char *string);                     // Decode pinyin.
static unsigned int get_hex (char *data);                       // Get hex value
static int   get_int        (char *data);                       // Get int value
static int   get_kanjicode  (char *data);                       // Extract JWPce compressed code from buffer
static char *get_param      (char *ptr);                        // Extract parameter from string.
static int   jis_to_code    (int jis);                          // Convert JIS code to JWPce compressed code.
static void  write_block    (FILE *file,byte *byffer,int count);// Write binary block to file.
static void  write_string   (FILE *file,char *string);          // Write string to file
static void  write_kana     (FILE *file,byte *kana);            // Write kana byte to file.

//--------------------------------
//
//  Adds a byte to the extneded data array.
//
//      buffer -- Buffer to add to.
//      pos    -- Current position, will be modified.
//      value  -- Value to be written
//
static void add_byte (byte *buffer,int &pos,int value) {
  buffer[pos++] = value;
  return;
}

//--------------------------------
//
//  Adds a code value (really just a byte) to the extneded data array.
//
//      buffer -- Buffer to add to.
//      pos    -- Current position, will be modified.
//      value  -- Value to be written
//
static void add_code (byte *buffer,int &pos,int value) {
  buffer[pos++] = value;
  return;
}

//--------------------------------
//
//  Adds the most common entry in one shot.  This is simply add a code and a 
//  short interger value.
//
//      buffer -- Buffer to add to.
//      pos    -- Current position, will be modified.
//      code   -- Code for value.
//      value  -- Value to be written
//
static void add_entry (byte *buffer,int &pos,int code,int value) {
  add_code  (buffer,pos,code);
  add_short (buffer,pos,value);
  return;
}

//--------------------------------
//
//  Adds a long integer to the extneded data array.
//
//      buffer -- Buffer to add to.
//      pos    -- Current position, will be modified.
//      value  -- Value to be written
//
static void add_long (byte *buffer,int &pos,long value) {
  *((long *) &buffer[pos]) = value;
  pos += 4;
  return;
}

//--------------------------------
//
//  Adds a short unsigned integer to the extneded data array.
//
//      buffer -- Buffer to add to.
//      pos    -- Current position, will be modified.
//      value  -- Value to be written
//
static void add_short (byte *buffer,int &pos,unsigned int value) {
  *((unsigned short *) &buffer[pos]) = value;
  pos += 2;
  return;
}

//--------------------------------
//
//  Convert from the code used by JWPce backward to the JIS code.
//
//      code   -- JWPce kanji code (compressed code).
//
//      RETURN -- JIS code.
//
static int code_to_jis (int code) {
  int temp,jis_hi,jis_lo;
  temp   = code/94;
  jis_hi = temp+0x30;
  jis_lo = (code-temp*94)+0x21;
  return ((jis_hi << 8) | jis_lo);
}

//--------------------------------
//
//  This routine decodes a pin-yin entry from the kanji-info file.  Note,
//  that the entry is decoded in place, modifing the argument string.
//
//  These entries are encoded as a character string, followed by a number
//  indicating the tone to be produced.  They are translated to a 
//  character string without the following number, where the inflection
//  over the vowel indicates the inflection.
//
//      string -- Input string.
//
//      return -- Pointer to the input string, which has been modifed.
//
static char *decode_pinyin (char *string) {
  int i,j,tone = 0;
    static char vmap[][4] = { { 'a',(char) 225,(char) 226,(char) 224 },     // Tonal vowels.
                              { 'e',(char) 233,(char) 234,(char) 232 },
                              { 'i',(char) 237,(char) 238,(char) 236 },
                              { 'o',(char) 243,(char) 244,(char) 242 },
                              { 'u',(char) 250,(char) 251,(char) 249 } };

  if (!(i = strlen(string))) return (string);               // No string, exit.
  for (j = 0; j < i; j++) string[j] = tolower(string[j]);   // Convert all to lower case.
  i--;                                                      // Ready to remove last char.
  if (('0' <= string[i]) && ('9' >= string[i])) {           // Get tone value.
    tone = string[i]-'0';
    string[i] = 0;
  }
  if ((tone <= 0) || (tone > 4)) return (string);           // Tone zero is unmodified case.
  for (j = 0; j < i; j++) {
    switch (string[j]) {                                    // Find vowel, modify, and exit.
      case 'a': string[j] = vmap[0][tone-1]; return (string);
      case 'e': string[j] = vmap[1][tone-1]; return (string);
      case 'i': string[j] = vmap[2][tone-1]; return (string);
      case 'o': string[j] = vmap[3][tone-1]; return (string);
      case 'u': string[j] = vmap[4][tone-1]; return (string);
    }
  }
  return (string);
}

//--------------------------------
//
//  Get a hex value from a string.
//
//      data   -- String to extract value from.
//
//      RETURN -- Integer value obtained.
//
static unsigned int get_hex (char *data) {
  unsigned int value;
  sscanf (data,"%x",&value);
  return (value);
}

//--------------------------------
//
//  Get an int value from a string.
//
//      data   -- String to extract value from.
//
//      RETURN -- Interger value obtained.
//
static int get_int (char *data) {
  int value;
  sscanf (data,"%d",&value);
  return (value);
}

//--------------------------------
//
//  Gets the kanjicode from the current line.  This is necessary, 
//  becasue this defines our index into the kanji array.  This needs
//  to be done because there is a gap in the database because there 
//  is a gap in the kanji between the type I and type II kanji.
//
//      data   -- Line to extract the kanjicode from.
//
//      RETURN -- The kanji code.  This is the code by which
//                JWPce will access the character.
//
static int get_kanjicode (char *data) {
  char *ptr,buffer[1000];
  if (*data == '#') {                   // Not a real character.  This is the file ID line.
    for (ptr = data; *ptr && (*ptr != '.'); ptr++);
    if (*ptr) ptr++;
    *ptr = 0;
    printf ("  %s\n",data+2);
    return (-1);
  }
  strcpy (buffer,data);                 // Copy string because parsing will mangle it.
  data = get_param(buffer);             // Skip character data
  get_param (data);                     // Frame data we want
  return (jis_to_code(get_hex(data)));  // return code.
}

//--------------------------------
//
//  Isolate a parameter from a string.
//
//      ptr    -- Pointer to the beginning of the string.
//  
//      RETURN -- Pointer to next parameter in the stirng or to the NULL.
//
static char *get_param (char *ptr) {
  if (*ptr == '{') {
    while (*ptr && (*ptr != '}')) ptr++;
    if (*ptr) { *ptr = 0; ptr++; }
  }
  else {
    while (*ptr && (*ptr != ' ')) ptr++;
  }
  if (*ptr) { *ptr = 0; ptr++; }
  return (ptr);
}

//--------------------------------
//
//  Convert from JIS code to compressed kanji code used by JWPce to 
//  access the kanji array.  The compressed code simply maps the character
//  codes into a continugous sequence of number.
//
//      jis    -- JIS code.
//
//      RETURN -- Compressed kanji code used by JWPce.
//
static int jis_to_code (int jis) {
  int code;
  jis -= 0x3021;
  code = (jis >> 8)*94+(jis & 0xff);
  return (code);
}

//--------------------------------
//
//  Write a block to the file.  The contents of the block can be anything.
//
//      file   -- File to write to.
//      buffer -- Buffer to write
//      count  -- Number of bytes to write.
//
static void write_block (FILE *file,byte *buffer,int count) {
  fwrite (buffer,count,1,file);
  offset += count+1;
  return;
}

//--------------------------------
//
//  Write a kana string to the file.  This is used to write a kana
//  string to the variable part of the file.  Note the strings 
//  written by this routine are stripped of the kana-base, and just
//  contain one byte of data.
//
//  The flollowing special codes are supported:
//
//      . -- Indicates the okurigana.  In the written file, the high
//           bit is set on the first of the okurigana characters.
//      - -- Indicates a dash, but is encoded in the file as 0x20 (space).
//
//      file -- File to write to.
//      kana -- Kana string in EUC.
//
static void write_kana (FILE *file,byte *kana) {
  int i;
  byte buffer[100];
  i = 0;
  for (i = 0; *kana; kana++) {
    if (*kana & 0x80) {
      buffer[i++] = *++kana & 0x7f;
    }
    else if (*kana == '.') {
#if 0                                   // Old encoding not used any more!
      buffer[i++] = 0x20;               // Encode okurigana with 0x20
#else
      kana += 2;
      buffer[i++] = *kana | 0x80;       // Encode okurigana with high-bit
#endif
    }
    else if (*kana == '-') {
      buffer[i++] = 0x1f;
    }
    else {
      printf ("Error unexpected character in kun-write\n");
      exit   (5);
    }
  }
  buffer[i] = 0;
  write_string (file,(char *) buffer);
  return;      
}

//--------------------------------
//
//  Write a string to the file.  This is a string in the varialbe 
//  part of the file.  This will include a null terminator, and will
//  correct the offest variable.
//
//      file   -- File to write to.
//      string -- String to write
//
static void write_string (FILE *file,char *string) {
  fwrite (string,strlen(string)+1,1,file);
  offset += strlen(string)+1;
  return;
}

//===================================================================
//
//  Class kinfo_big.
//

//--------------------------------
//
//  Generate an error message about this kanji.
//
void kinfo_big::error (char *text) {
  printf ("KANJI ERROR: JIS=%x  err=%s\n",jis,text);
  return;
}

//--------------------------------
//
//  This rotuine parcess a string from Jim Breens KANJIDIC and 
//  converts it into a structre stored here.
//
//      data -- String from KANJIDIC.
//
int kinfo_big::parse (char *data) {
  short nanori = false;
  short special = false;
  char *ptr,*ptr2,*p;
  byte *bptr;
  memset (this,0,sizeof(kinfo_big));    // Intialize the structure to zero.
  fc_main  = -1;                        // Because 0000.0 is a vaid code we must use -1 as no code.
  fc2_main = -1;
  ptr      = get_param(data);
  ptr2     = get_param(ptr);
  jis      = get_hex(ptr);
  ptr      = ptr2;
  while (*ptr) {
    ptr2 = get_param(ptr);
    switch (*ptr) {
      case 'B': // Bushu
           if (bushu) error ("Two bushu entries");
           bushu = get_int(ptr+1);
           break;
      case 'C': // Classical Bushu
           if (classical) error ("Two classical bushu entires");
           classical = get_int(ptr+1);
           break;
      case 'D': // Dictionary codes
           switch (ptr[1]) {
             case 'B':  // Japanese For Busy People vols I-III, published by the AJLT. The codes are the volume.chapter.                  
                  if (busy_chapter) error ("Two 'Japanese For Busy People' vols I-III, published by the AJLT dictionary codes");
                  busy_vol = ptr[2]-'0';
                  if ((busy_vol < 1) || (busy_vol > 3)) error ("Invalid volume for Japanese For Busy People dictionary entry");
                  if (ptr[3] != '.') error ("Dot expected in Japanese For Busy People dictionary entry");
                  busy_chapter = get_int(ptr+4);
                  break;
             case 'C':  // The Kanji Way to Japanese Language Power by Dale Crowley.
                  if (easy_way) error ("Two 'The Kanji Way to Japanese Language Power' by Dale Crowley dictionary entires");
                  easy_way = get_int(ptr+2);
                  break;
             case 'G':  // Kodansha Compact Kanji Guide
                  if (compact_guide) error ("Two Kodansha Compact Kanji Guide entries");
                  compact_guide = get_int(ptr+2);
                  break;
             case 'J':  // Kanji in Context by Nishiguchi and Kono                  
                  if (kanji_context) error ("'Two Kanji in Context' by Nishiguchi and Kono dictionary entires");
                  kanji_context = get_int(ptr+2);
                  break;
             case 'K':  // Jack Halpern in his Kanji Learners Dictionary, published by Kodansha in 1999
                  if (halpern_kld) error ("Two Halpern Kanji Learners Dictionary entires");
                  halpern_kld = get_int(ptr+2);
                  break;
             case 'R':  // DeRoo codes
                  if (deroo) error ("Two DeRoo entires");
                  deroo = get_int(ptr+2);
                  break;
             case 'O':  // P.G. O'Neill's Essential Kanji 
                  if (oneill_ek) error ("Two P.G. O'Neill's Essential Kanji entires");
                  oneill_ek = get_int(ptr+2);
                  break;
             case 'S':  // A Guide To Reading and Writing Japanese" edited by Florence Sakade.
                  if (read_write) error ("Two 'A Guide To Reading and Writing Japanese' edited by Florence Sakade dictionary entries");
                  read_write = get_int(ptr+2);
                  break;
             case 'T':  // Tuttle Kanji Cards, compiled by Alexander Kask
                  if (tuttle_card) error ("Two Tuttle Kanji Card dictionary entries");
                  tuttle_card = get_int(ptr+2);
                  break;
             default:
                  printf ("Unknown dictionary code '%c' [%x %d]\n",ptr[1],jis,count); 
                  break;
           }
           break;
      case 'E': // "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
           if (henshall) error ("Two 'A Guide To Remembering Japanese Characters' by Kenneth G. Henshall entries");
           henshall = get_int(ptr+1);
           break;
      case 'F': // Frequency
           if (freq) error ("Two frequency entries");
           freq = get_int(ptr+1);
           break;
      case 'G': // Grade
           if (grade) error ("Two grade entries");
           grade = get_int(ptr+1);
           break;
      case 'H': // Halpern index
           if (halpern) error ("Two Halpern entries");
           halpern = get_int(ptr+1);
           break;
      case 'I': // Spahn & Hadamitzky. 
           if (ptr[1] == 'N') {
             if (sh_kanakanji) error ("Two Spahn and Hadamitzky Kana & Kanji entries");
             sh_kanakanji = get_int(ptr+2);
           }
           else {
             char *pp;
             if (sh_rstroke || sh_ostroke) error ("Two Spahn and Hadamitsky Dictionary entries");
             for (p = ptr+2; isdigit(*p); p++);
             sh_radical = *p;
             *p         = ' ';
             sh_rstroke = get_int(ptr+1);
             for (pp = p+1; *pp != '.'; pp++);
             *pp = ' ';
             sh_ostroke = get_int(p+1);
             sh_index   = get_int(pp+1);
           }
           break;
      case 'K': // Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
           if (gakken) error ("Two Gakken Kanji Dictionary entries");
           gakken = get_int(ptr+1);
           break;
      case 'L': // "Remembering The Kanji" by James Heisig
           if (heisig) error ("Two 'Remembering The Kanji' by James Heisig entries");
           heisig = get_int(ptr+1);
           break;
      case 'M': // Morohashi Daikanwajiten
           if (ptr[1] == 'N') {
             if (md_long) error ("Two Morohashi Daikanwajiten long entries");
             for (p = ptr+2; isdigit(*p); p++);
             if (*p == 'P') md_mark = *p;
             if (*p == 'X') md_mark = *p;
             *p = ' ';
             md_long = get_int(ptr+2);
           }
           else if (ptr[1] == 'P') {
             if (md_short1) error ("Two Morohashi Daikanwajiten short entries");
             for (p = ptr+2; isdigit(*p); p++);
             *p = ' ';
             md_short1 = get_int(ptr+2);
             md_short2 = get_int(p+1);
           }
           else {
             error ("Invalid Morohashi Daikanwajiten code (not P or N)");
           }
           break;
      case 'N': // Nelson index
           if (nelson) error ("Two Nelson entries");
           nelson = get_int(ptr+1);
           break;
      case 'O': // "Japanese Names", by P.G. O'Neill
           if (oneill) error ("Two 'Japanese Names', by P.G. O'Neill entries");
           oneill = get_int(ptr+1);
           break;
      case 'P': // skip code
           ptr[2] = 0;
           for (p = ptr+3; *p != '-'; p++);
           *p++ = 0;
           skip_type = get_int(ptr+1);
           skip_1    = get_int(ptr+3);
           skip_2    = get_int(p);
           if (!skip_type || !skip_1 || !skip_2) error ("Skip code has zeros");
           break;
      case 'Q': // Four Corner
           if (!strncmp(ptr,"Q0.0",4)) {
             error ("Ivalid four corners code 0.0");
             break;
           }
           if (fc2_main != -1) error ("Three Four Corners entries");
           if (ptr[5] != '.') error ("Four Corner codes does not have a dot");
           ptr[5] = ' ';           
//           if (!get_int(ptr+1)) error ("Four Corner code of 0000 found");
           if (fc_main != -1) {
             fc2_main  = get_int(ptr+1);
             fc2_index = ptr[6]-'0';
           }
           else {
             fc_main  = get_int(ptr+1);
             fc_index = ptr[6]-'0';
           }
           break;
      case 'S': // Stroke count
           if (!stroke) stroke = get_int(ptr+1);    // More than first entry are mis-readings
             else {
               int i;
               for (i = 0; (i < NUMBER_STROKES) && strokes[i]; i++);
               if (i == NUMBER_STROKES) { printf ("Too many stroke miscounts [%x %d]\n",jis,i); return (true); }
               strokes[i] = get_int(ptr+1);
             }
           break;
      case 'T': // Makrer indicating nanori and other special entries.
           switch (ptr[1]) {
             case '1': nanori = true; break;
             default: special = true; nanori = false; break;
           }
           break;
      case 'U': // Unicode
           if (unicode) error ("Two UNICODE entries");
           unicode = get_hex(ptr+1);
           break;
      case 'V': // New Nelson index (edited by Haig)
           if (haig) error ("Two Haig (New Nelson) entries");
           haig = get_int(ptr+1);
           break;
      case 'W': // Korean
           for (p = ptr+1; !isspace((byte) *p); p++);               // Byte cast is required because of bug in VC++ debugging rotuines.
           *p = 0;
           if (strlen(korean)+strlen(ptr+1)+2 > SIZE_KOREAN) { printf ("Korean string is too long [%x %d]\n",jis,count); return (true); }
           if (korean[0]) strcat (korean,", ");
           strcat (korean,ptr+1);
           *p = ' ';
           break;
      case 'Y': // Pinyin
           p = decode_pinyin(ptr+1);
           if (strlen(pinyin)+strlen(p)+2 > SIZE_PINYIN) { printf ("PinYin string is too long [%x %d]\n",jis,count); return (true); }
           if (pinyin[0]) strcat (pinyin,", ");
           strcat (pinyin,p);
           break;
      case '{': // Meaning entries.
           if (strlen(ptr+1) >= SIZE_IMI) { printf ("imi string is too long [%x %d]\n",jis,count); return (true); }
           strcpy (imi[imi_count++],ptr+1);
           if (imi_count == NUMBER_IMI) { printf ("Too many imi-readings [%x %d]\n",jis,count); return (true); }
           break;
      case 'X': // Cross-refference code
           switch (ptr[1]) {
             case 'B':
                  if (x_bushu) error ("Two bushu cross-reference entries");
                  x_bushu = get_int(ptr+2);
                  break;
             case 'D':
                  switch (ptr[2]) {
                    case 'R':
                         if (x_deroo2) error ("Three DeRoo cross-reference entries");
                         if (x_deroo) x_deroo2 = get_int(ptr+3); else x_deroo = get_int(ptr+3);
                         break;
                    default:
                         printf ("Unknown dictionary cross-reference code of %c [%x %d]\n",ptr[2],jis,count); 
                         return (true); 
                  }
                  break;
             case 'H':
                  if (x_halpern2) error ("Three Halpern cross-reference entries");
                  if (x_halpern) x_halpern2 = get_int(ptr+2); else x_halpern = get_int(ptr+2);
                  break;
             case 'I': {
                    char *pp;
                    if (x_shrstroke || x_shostroke) error ("Two Spahn and Hadamitsky Dictionary cross-reference entries");
                    for (p = ptr+3; isdigit(*p); p++);
                    x_shradical = *p;
                    *p          = ' ';
                    x_shrstroke = get_int(ptr+2);
                    for (pp = p+1; *pp != '.'; pp++);
                    *pp = ' ';
                    x_shostroke = get_int(p+1);
                    x_shindex   = get_int(pp+1);
                  }
                  break;
             case 'J':
                  switch (ptr[2]) {
                    case '0':
                         if (x_jis0208c == NUMBER_JISXREF) { error ("Too many JIS 0208 cross-reference entries"); break; }
                         x_jis0208[x_jis0208c++] = get_hex(ptr+3);
                         break;
                    case '1':
                         if (x_jis0212c == NUMBER_JISXREF) { error ("Too many JIS 0212 cross-reference entries"); break; }
                         x_jis0212[x_jis0212c++] = get_hex(ptr+3);
                         break;
                    default:
                         printf ("Unkown coding is JIS cross-reference '%c' [%x %d]\n",ptr[2],jis,count);
                         break;
                  }
                  break;
             case 'N':
                  if (x_nelson2) error ("Three Crossreference-Nelson entries");
                  if (x_nelson) x_nelson2 = get_int(ptr+2); else x_nelson = get_int(ptr+2);
                  break;
             case 'O':
                  if (x_oneill) error ("Two O'Neill cross-reference entries");
                  x_oneill = get_int(ptr+2);
                  break;
             default:
                  printf ("Unknown Cross-refference code '%c' [%x %d]\n",ptr[1],jis,count); 
           }
           break;
      case 'Z': // Miss-classification code
           if (z_skipc == NUMBER_SKIPXREF) { error ("Too many SKIP miss-classification codes"); break; }
           z_skipe[z_skipc  ] = ptr[1];
           while (!isdigit(*ptr)) ptr++;
           for (p = ptr+2; *p != '-'; p++);
           z_skipt[z_skipc  ] = get_int (ptr);
           z_skip1[z_skipc  ] = get_int (ptr+2);
           z_skip2[z_skipc++] = get_int (p+1);
           break;
      case ' ':
      case '\r':
      case '\n':
           break;
      default:  // These are unused entries, and the kun/on readings.
           bptr  = (byte *) ptr;
           if (*ptr == '-') bptr++;
           if (!(*bptr & 0x80)) {
             printf ("Unknown entry code '%c' [%x %d]\n",*ptr,jis,count); 
             break;
           }
           if (*bptr == 0xA5) {   // on readings.
             if (strlen(ptr+1) >= SIZE_ON) { printf ("on string is too long [%x %d]\n",jis,count); return (true); }
             strcpy ((char *) on[on_count++],ptr);
             if (on_count == NUMBER_ON) { printf ("Too many on-readings [%x %d]\n",jis,count); return (true); }
           }
           else if (nanori) {     // nanori readings
             if (strlen(ptr+1) >= SIZE_NAN) { printf ("nan string is too long [%x %d]\n",jis,count); return (true); }
             strcpy ((char *) nan[nan_count++],ptr);
             if (nan_count == NUMBER_NAN) { printf ("Too many nanori-readings [%x %d]\n",jis,count); return (true); }
           }
           else if (special) {    // Unused entries T2...
           }
           else {                 // kun readings.
             if (strlen(ptr+1) >= SIZE_KUN) { printf ("kun string is too long [%x %d]\n",jis,count); return (true); }
             strcpy ((char *) kun[kun_count++],ptr);
             if (kun_count == NUMBER_KUN) { printf ("Too many kun-readings [%x %d]\n",jis,count); return (true); }
           }
           break;
    }
    ptr  = ptr2;
  }
  count++;
  return (false);
}

//--------------------------------
//
//  Write this element to the file.  This routine assumes that the 
//  file is in the correct location for the fixed structure write.
//  The routine will seek to the correct location for the variable 
//  data writes.
//
//      file -- File to write to.
//
#define ADD_CODE(x)     *((ushort *) (extra)) |= x

void kinfo_big::write (FILE *file,int extended) {
  int i,size;
  kinfo   info;
  extend *extend;
  byte    extra[500];
//
//  Build the endtend data structure.
//
  memset (extra,0,sizeof(extra));
  extend = (struct extend *) extra;
  extend->fc_main    = fc_main;
  extend->fc_index   = fc_index;
  extend->fc_index2  = fc2_index;
  extend->md_long    = (ushort) md_long;
  extend->md_short1  = md_short1;
  extend->md_short2  = md_short2;
  extend->md_p       = md_mark == 'P';
  extend->md_x       = md_mark == 'X';
  extend->sh_rstroke = sh_rstroke;
  extend->sh_radical = sh_radical-'a';
  extend->sh_ostroke = sh_ostroke;
  extend->sh_index   = sh_index;
  size = sizeof(struct extend);
//
//  Build the addtional string.
//
  if (busy_chapter  ) add_entry (extra,size,'B',(busy_vol << 8) | busy_chapter);   // Japanese For Busy People vols I-III, published by the AJLT. The codes are the volume.chapter.
  if (easy_way      ) add_entry (extra,size,'C',easy_way);      // The Kanji Way to Japanese Language Power by Dale Crowley.
  if (deroo         ) add_entry (extra,size,'D',deroo);         // DeRoo entry for kanji
  if (henshall      ) add_entry (extra,size,'E',henshall);      // "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
  if (freq          ) add_entry (extra,size,'F',freq);          // Frequency count
  if (compact_guide ) add_entry (extra,size,'G',compact_guide); // Kodansha Compact Kanji Guide
  if (halpern_kld   ) add_entry (extra,size,'H',halpern_kld);   // Jack Halpern in his Kanji Learners Dictionary, published by Kodansha in 1999
  if (sh_kanakanji  ) add_entry (extra,size,'I',sh_kanakanji);  // Spahn and Hadamitzky Kana & Kanji
  if (kanji_context ) add_entry (extra,size,'J',kanji_context); // Kanji in Context by Nishiguchi and Kono                  
  if (gakken        ) add_entry (extra,size,'K',gakken);        // Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
  if (heisig        ) add_entry (extra,size,'L',heisig);        // "Remembering The Kanji" by James Heisig
  if (oneill_ek     ) add_entry (extra,size,'N',oneill_ek);     // P.G. O'Neill's Essential Kanji 
  if (oneill        ) add_entry (extra,size,'O',oneill);        // "Japanese Names", by P.G. O'Neill
  if (read_write    ) add_entry (extra,size,'S',read_write);    // A Guide To Reading and Writing Japanese" edited by Florence Sakade.
  if (tuttle_card   ) add_entry (extra,size,'T',tuttle_card);   // Tuttle Kanji Cards, compiled by Alexander Kask
  if (fc2_main != -1) add_entry (extra,size,'Q',fc2_main);      // Four-corners secondary code
//
//  Cross reference entries.
//
  if (x_halpern2) add_entry (extra,size,'h',x_halpern2);
  if (x_halpern ) add_entry (extra,size,'h',x_halpern);
  if (x_nelson2 ) add_entry (extra,size,'n',x_nelson2);
  if (x_nelson  ) add_entry (extra,size,'n',x_nelson);
  if (x_oneill  ) add_entry (extra,size,'o',x_oneill);
  if (x_deroo   ) add_entry (extra,size,'d',x_deroo);
  for (i = 0; i < x_jis0208c; i++) add_entry (extra,size,'k',x_jis0208[i]);
  for (i = 0; i < x_jis0212c; i++) add_entry (extra,size,'j',x_jis0212[i]);
  for (i = 0; i < z_skipc; i++) {
    int j;
    switch (z_skipe[i]) {
      case 'P': j = SKIP_POSITION; break;
      case 'S': j = SKIP_STROKE;   break;
      case 'B': j = SKIP_BOTH;     break;
      case 'R': j = SKIP_BREEN;    break;
      default:
           error ("Unknown skip miss-clasification type");
           break;
    }
    j = (j << 13) | (z_skipt[i] << 10) | (z_skip1[i] << 5) | z_skip2[i];
    add_entry (extra,size,'z',j);
  }
//
//  Finish up
//
  size++;
  if ((size == sizeof(struct extend)+1) && (fc_main == -1) && !md_long && !md_short1 && !sh_rstroke && !sh_ostroke) size = 0;
  if (!stroke) size = 0;
  if (!jis) size = 0;
  if (!extended) size = 0;              // Disable the extended fields
//
//  Do the fixed part
//
  info.bushu     = bushu;                   // Fixed part.
  info.classical = classical;
  info.strokes   = stroke;
  info.skip_t    = skip_type;
  info.skip_1    = skip_1;
  info.skip_2    = skip_2;
  info.pinyin    = pinyin[0] ? 1 : 0;
  info.on        = on_count;
  info.kun       = kun_count;
  info.imi       = imi_count;
  info.nan       = nan_count;
  info.halpern   = halpern;
  info.nelson    = nelson;
  info.haig      = haig;
  info.offset    = offset;
  info.grade     = grade;
  info.korean    = korean[0] ? 1 : 0;
  info.extra     = (size > 2) ? 1 : 0;
  fwrite (&info,sizeof(kinfo),1,file);
//
//  Do the varialbe part
//
  fseek (file,offset,SEEK_SET);             // Variable parts.
  if (korean[0]) write_string (file,korean);
  if (pinyin[0]) write_string (file,pinyin);
  for (i = 0; i < imi_count; i++) write_string (file,imi[i]);
  for (i = 0; i < on_count ; i++) write_kana   (file,on [i]);
  for (i = 0; i < kun_count; i++) write_kana   (file,kun[i]);
  for (i = 0; i < nan_count; i++) write_kana   (file,nan[i]);
  if (size) write_block (file,extra,size);
  return;
}

//--------------------------------
//
//  Calaculate the byte system of the variable length items in an entry.
//
//      RETURN -- byte-size;
//
int kinfo_big::size () {
  int i,j;
  int size;
  size  = strlen(pinyin)+1;
  size += strlen(korean)+1;
  for (i = 0; i < imi_count; i++) size += strlen(imi[i])+1;
  for (i = 0; i < on_count ; i++) size += strlen((char *) on [i])/2+1; 
  for (i = 0; i < nan_count; i++) size += strlen((char *) nan[i])/2+1;
  for (i = 0; i < kun_count; i++) {
    for (j = 0; kun[i][j]; j++) {
      if (kun[i][j] == '.') ;
      else if (kun[i][j] == '-') size++;
      else { size++; j++; }
    }
  }
  return (size);
}

//===================================================================
//
//  High level process routines.
//

//--------------------------------
//
//  Read the KANJIDIC data base.
//
//      name -- Name of database file.
//
static int read_database (char *name,int encode) {
  int   i,j;
  FILE *file;
  char buffer[5000];
  printf ("Reading kanji data base [%s] ... \n",name);
  if (!(file = fopen(name,"rb"))) { printf ("\nCannot open database\n"); return (true); }
  memset (kanji,0,sizeof(kanji));
  while (true) {
    fgets (buffer,sizeof(buffer),file);
    if (feof(file)) break;
    if (translate(buffer,encode)) { printf ("Error -- Invalid character found (after JIS %x)\n\n",i); exit (5); }
    if (-1 != (i = get_kanjicode(buffer))) {            // There is a gap inthe kanji between set I and set II
      if (kanji[i].parse (buffer)) { fclose (file); return (true); }
    }
  }
  j = count;
  for (i = 0; i < NUMBER_BIGKANJI; i++) if (kanji[i].jis) count = i+1;
  printf ("done [%d kanji read (last JIS=%x)]\n",j,kanji[count-1].jis);
  fclose (file);
  return (false);
}

//--------------------------------
//
//  Print statistics on the size of various items.
//
static void print_stats () {
  int i,j;
  int on,kun,imi,nan,size,pinyin,korean,skip1,skip2,nohalpern,nonelson,maxhalpern,maxnelson,nohaig,maxhaig,maxstrokeerr;
  int nan_space,pinyin_space,stroke,korean_space;
  int sh_index,sh_kanakanji,sh_ostroke,sh_radical,sh_rstroke,no_sh;
  int md_long,md_short1,md_short2,no_mdlong,no_mdshort,max_busy;
  int no_fc,no_gakken,gakken,heisig,no_heisig,oneill,no_oneill,no_deroo,no_oneill_ek,no_halpern_kld;
  int no_kanjicontext,no_busypeople,no_compactguide,no_easyway,no_tuttlecard,no_readwrite;
  on = kun = imi = nan = size = pinyin = korean = skip1 = skip2 = nohalpern = nonelson = maxhalpern = maxnelson = nohaig = maxhaig = maxstrokeerr = 0;
  sh_index = sh_kanakanji = sh_ostroke = sh_radical = sh_rstroke = no_sh = 0;
  md_long = md_short1 = md_short2 = no_mdlong = no_mdshort = 0;
  no_fc = gakken = no_gakken = heisig = no_heisig = oneill = no_oneill = no_deroo = no_oneill_ek = no_halpern_kld = 0;
  no_kanjicontext = no_busypeople = no_compactguide = no_easyway = no_tuttlecard = no_readwrite = max_busy = 0;


  nan_space = pinyin_space = korean_space = stroke = 0;
  for (i = 0; i < count; i++) {
    if (kanji[i].skip_1 > skip1) skip1 = kanji[i].skip_1;
    if (kanji[i].skip_2 > skip2) skip2 = kanji[i].skip_2;
    if (kanji[i]. on_count > on ) on  = kanji[i]. on_count;
    if (kanji[i].kun_count > kun) kun = kanji[i].kun_count;
    if (kanji[i].imi_count > imi) imi = kanji[i].imi_count;
    if (kanji[i].nan_count > nan) nan = kanji[i].nan_count;
    if (kanji[i].stroke > stroke) stroke = kanji[i].stroke;
    if (strlen(kanji[i].pinyin) > (unsigned) pinyin) pinyin = strlen(kanji[i].pinyin);
    if (strlen(kanji[i].korean) > (unsigned) korean) korean = strlen(kanji[i].korean);
    if (kanji[i].size() > size) size = kanji[i].size();
    if (!kanji[i].nelson       ) nonelson++;
    if (!kanji[i].halpern      ) nohalpern++;
    if (!kanji[i].haig         ) nohaig++;
    if (!kanji[i].fc_main      ) no_fc++;
    if (!kanji[i].sh_rstroke   ) no_sh++;
    if (!kanji[i].md_long      ) no_mdlong++;
    if (!kanji[i].md_short1    ) no_mdshort++;
    if (!kanji[i].gakken       ) no_gakken++;
    if (!kanji[i].heisig       ) no_heisig++;
    if (!kanji[i].oneill       ) no_oneill++;
    if (!kanji[i].deroo        ) no_deroo++;
    if (!kanji[i].oneill_ek    ) no_oneill_ek++;
    if (!kanji[i].halpern_kld  ) no_halpern_kld++;
    if (!kanji[i].kanji_context) no_kanjicontext++;
    if (!kanji[i].busy_chapter ) no_busypeople++;
    if (!kanji[i].compact_guide) no_compactguide++;
    if (!kanji[i].easy_way     ) no_easyway++;
    if (!kanji[i].tuttle_card  ) no_tuttlecard++;
    if (!kanji[i].read_write   ) no_readwrite++;
    if (kanji[i].nelson  > maxnelson ) maxnelson  = kanji[i].nelson;
    if (kanji[i].halpern > maxhalpern) maxhalpern = kanji[i].halpern;
    if (kanji[i].haig    > maxhaig   ) maxhaig    = kanji[i].haig;
    if (kanji[i].pinyin[0]) pinyin_space += strlen(kanji[i].pinyin)+1;
    if (kanji[i].korean[0]) korean_space += strlen(kanji[i].korean)+1;
    if (kanji[i].sh_index     > sh_index    ) sh_index     = kanji[i].sh_index;
    if (kanji[i].sh_kanakanji > sh_kanakanji) sh_kanakanji = kanji[i].sh_kanakanji;
    if (kanji[i].sh_ostroke   > sh_ostroke  ) sh_ostroke   = kanji[i].sh_ostroke;
    if (kanji[i].sh_radical   > sh_radical  ) sh_radical   = kanji[i].sh_radical;
    if (kanji[i].sh_rstroke   > sh_rstroke  ) sh_rstroke   = kanji[i].sh_rstroke;
    if (kanji[i].md_long      > md_long     ) md_long      = kanji[i].md_long;
    if (kanji[i].md_short1    > md_short1   ) md_short1    = kanji[i].md_short1;
    if (kanji[i].md_short2    > md_short2   ) md_short2    = kanji[i].md_short2;
    if (kanji[i].gakken       > gakken      ) gakken       = kanji[i].gakken;
    if (kanji[i].heisig       > heisig      ) heisig       = kanji[i].heisig;
    if (kanji[i].oneill       > oneill      ) oneill       = kanji[i].oneill;
    if (kanji[i].busy_chapter > max_busy    ) max_busy     = kanji[i].busy_chapter;
    for (j = 0; j < kanji[i].nan_count; j++) nan_space += strlen((char *) kanji[i].nan[j])/2+1;
    for (j = 0; (j < NUMBER_STROKES) && kanji[i].strokes[j]; j++);      // Determine the number of stroke mis-counts
    if (j > maxstrokeerr) maxstrokeerr = j;
  }
  printf ("Statistics:\n\n");
  printf ("  MAX OF TYPE: on=%d  kun=%d  imi=%d  nan=%d\n",on,kun,imi,nan);
  printf ("  MAX VALUES : skip1=%d  skip2=%d  strokes=%d\n",skip1,skip2,stroke);
  printf ("               halpern=%d  nelson=%d  haig=%d  stroke-err=%d  Busy=%d\n",maxhalpern,maxnelson,maxhaig,maxstrokeerr,max_busy);
  printf ("  SPAN & HADA: Kana&Kanji=%d  rstr=%d rad=%c(%d) ostr=%d indx=%d\n",sh_kanakanji,sh_rstroke,sh_radical,sh_radical-'a',sh_ostroke,sh_index);
  printf ("  MOROHASHI  : long=%d  short=%d %d  Gakken=%d  Heisig=%d  ONeill=%d\n",md_long,md_short1,md_short2,gakken,heisig,oneill);
  printf ("  NO DICT CHK: table=%d  Halpern=%d  Nelson=%d  Haig=%d  SH=%d  4C=%d\n",count,nohalpern,nonelson,nohaig,no_sh,no_fc);  
  printf ("               MDl=%d  MDs=%d  Gakken=%d  Heisig=%d  ONeill=%d\n",no_mdlong,no_mdshort,no_gakken,no_heisig,no_oneill);
  printf ("               DeRoo=%d  ONeill(EK)=%d  Halpern(KLD)=%d  KanCon=%d\n",no_deroo,no_oneill_ek,no_halpern_kld,no_kanjicontext);
  printf ("               BusyPeople=%d  CompactGuide=%d  EasyWay=%d\n",no_busypeople,no_compactguide,no_easyway);
  printf ("               TuttleCard=%d  ReadWrite=%d\n",no_tuttlecard,no_readwrite);
  printf ("  SIZES(byte): pinyin(line)=%d  korean(line)=%d  variable=%d  struct=%d\n",pinyin,korean,size,sizeof(kinfo));
  printf ("  TOTAL SIZES: pinyin(total)=%d  korean(total)=%d  nanori(total)=%d\n\n",pinyin_space,korean_space,nan_space);
  return;
}

//--------------------------------
//
//  Write frequency of kanji's information.
//
//      name -- Name of file to write.
//
static int write_frequency (char *name) {
  FILE *file;
  int   i,j;
  KANJI jis;
  int   total = 0;
  printf ("Writing frequent kanji [%s] ... ",name);
  if (!(file = fopen(name,"w"))) { printf ("\nCannot open database\n"); return (true); }
  for (i = 1; i < 3000; i++) {                              // I counts through frequencies.
    for (j = 0; (j < count) && (kanji[j].freq != i); j++);  // Find the kanji with the frequency
    if (j == count) continue;                               // Used to abort here, but one frequency is missing.
    total++;                                                // Output the kanji.
    jis = (((kanji[j].jis & 0xff) << 8) | (kanji[j].jis >> 8)) | 0x8080;    // Swap the bytes.
    fwrite (&jis,2,1,file);
    if (!(i % 25)) fputs ("\n",file);                       // New line every 25 kanji.
    if (!(i % 100)) fprintf (file,"%d\n",i);                // New couter every 100 kanji.
  }
  fputs  ("\n",file);
  fclose (file);
  printf ("done [%d kanji output]\n",total);
  return (false);
}

//--------------------------------
//
//  Write the kanji-info database.
//
//      name     -- Name of file.
//      extended -- Write extended fields.
//      flags    -- This indicates what is in the file.
//
//      RETURN   -- A non-zero return indicates a write error.
//
static int write_kanji (char *name,int extended,int flags) {
  int i;
  ulong  magic  = KINFO_MAGIC;
  ushort fsize  = count;
  ushort maxjis = 0;
  ulong  fflag  = flags;
  FILE *file;
  offset = KANJIINFO_OFFSET+sizeof(kinfo)*count;
  printf ("Writing kanji database [%s] ... ",name);
  if (!(file = fopen(name,"wb"))) { printf ("\nCannot open database [%s]\n",name); return (true); }
//
//  Setup header information.
//
  for (i = 0; i < count; i++) {
    if (kanji[i].jis > maxjis) maxjis = kanji[i].jis;
  }
  fwrite (&magic ,sizeof(ulong ),1,file);
  fwrite (&fflag ,sizeof(ulong ),1,file);
  fwrite (&fsize ,sizeof(ushort),1,file);
  fwrite (&maxjis,sizeof(ushort),1,file);
//
//  Write the kanji data.
//
  for (i = 0; i < count; i++) {
    fseek (file,KANJIINFO_OFFSET+i*sizeof(kinfo),SEEK_SET);
    kanji[i].write (file,extended);
  }
  fseek  (file,0,SEEK_END);
  printf ("done [%d bytes]\n",ftell(file));
  fclose (file);
  return (false);
}

//--------------------------------
//
//  Write strke database.  This is a file used by the radical infomration
//  utility RINFO.EXE.  This file containsa a byte array of the stroke 
//  counts for the kanji organaized to be directly indexed based on the 
//  JIS code 
//
//      name   -- Name of data file to write.
//
//      RETURN -- A non-zero return indicates an error.
//  
static int write_stroke (char *name) {
  int    i;
  FILE  *file;
  static byte strokes[1 << 15];
  printf ("Writing Kanji Stroke Count Data File [%s] ... ",name);
  for (i = 0; i < sizeof(strokes); i++) strokes[i] = 0;
  for (i = 0; i < count; i++) strokes[kanji[i].jis] = (byte) kanji[i].stroke;
  if (!(file = fopen(name,"wb"))) { printf ("\nCannot open database [%s]\n",name); return (true); }
  fwrite (strokes,sizeof(strokes),1,file);
  fclose (file);
  printf ("done\n");
  return (false);
}

//--------------------------------
//
//  Writes a c source file that contains the main table for converting
//  kanji from JIS code to unicode.
//
//      name -- Name of file to write to.
//
static int write_unicode (char *name) {
  int    i,start;
  FILE  *file;
  printf ("Writing Unicode kanji conversion file [%s] ... ",name);

  if (!(file = fopen(name,"w"))) { printf ("\nCannot open outuput file [%s]\n",name); return (true); }
  fprintf (file,"  ");
  for (start = i = 0; i < count; i++) {
    fprintf (file,"0x%4.4x,",(kanji[i].unicode) & 0xffff);
    if (i && !(i % 10)) {
      fprintf (file,"  // 0x%4.4x..0x%4.4x\n  ",code_to_jis(start),code_to_jis(i));
      start = i+1;
    }
  }
  fclose (file);
  printf ("done\n");
  return (false);
}

//===================================================================
//
//  Main routine.
//

#define KIFLAG_LARGE    (KIFLAG_PINYIN | KIFLAG_KOREAN | KIFLAG_NANORI | KIFLAG_EXTRA | KIFLAG_EVAR | KIFLAG_XREF)
#define KIFLAG_MEDIUM   (KIFLAG_EXTRA | KIFLAG_EVAR | KIFLAG_XREF)
#define KIFLAG_SMALL    (0)

int main (int argc,char **argv) {
  int i;
  int   encode = ENCODE_EUC;
  int   stats  = false;
  char *input  = "KANJIDIC";
  printf ("KINFO Utility Verison 7.0 copyright (C) 1998-2002,2003 glenn rosenthal.\n");
  printf ("  Process Jim Breen's KANJIDIC file into JWPce KANJINFO.DAT file\n\n");
//
//  Process arguments
//
  for (i = 1; i < argc; i++) {
    if      (! stricmp (argv[i],"EUC"  )) encode = ENCODE_EUC;
    else if (! stricmp (argv[i],"UTF8" )) encode = ENCODE_UTF8;
    else if (! stricmp (argv[i],"STATS")) stats  = true;
    else if (!_strnicmp(argv[i],"IN=",3)) input  = argv[i]+3;
    else {
      printf ("ERROR -- Unknown parameter [%s]!\n\n     KINFO [ EUC ] [ UTF8 ] [ STATS ] [ IN=<name> ]\n\n",argv[i]);
      exit   (5);
    }
  }
//
//  The real work.
//
  if (read_database  (input,encode    )) { printf ("Error Reading Database\n"         ); exit (5); }
  if (write_stroke   ("kanjisrk.dat"  )) { printf ("Error Writing Kanji Storke File\n"); exit (5); }
//if (write_frequency("kanji_freq.euc")) { printf ("Error Writing Kanji Frequencies\n"); exit (5); }
  if (stats) print_stats ();
  if (write_kanji    ("kanjinfo.dat",true ,KIFLAG_LARGE)) { printf ("Error Writing Big kanji database\n"); exit (5); }
  for (i = 0; i < count; i++) kanji[i].nan_count = 0;
  for (i = 0; i < count; i++) kanji[i].pinyin[0] = 0;
  for (i = 0; i < count; i++) kanji[i].korean[0] = 0;
  if (write_kanji    ("kanjinfo.med",true ,KIFLAG_MEDIUM)) { printf ("Error Writing Medium kanji database\n"); exit (5); }
  if (write_kanji    ("kanjinfo.sml",false,KIFLAG_SMALL )) { printf ("Error Writing Small kanji database\n"); exit (5); }
  if (write_unicode  ("jwp_unic.dat"                    )) { printf ("Error Writing Unicdoe data file\n"); exit (5); }

  return (0);
}




