//-------------------------------------------------------------------//
//                                                                   //
//  KINFO Copyright (C) Glenn Rosenthal, 1998-2001,2002.             //
//  All rights reserved.                                             //
//                                                                   //   
//  The KANJIDIC file this program extracts information from was     //
//  compiled by Jim Breen, who holds the copyright on it.            //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  Structure of the KINFO files:
//
//  The following indicates the structure of the files written by this
//  utility.  The actual infirmaton content of the files can differ, but
//  the structure is identical.
//
//      MAGIC    -- 4 byte binary number used to verify the file type.
//      FLAGS    -- Flags indicating what is written into the file:
//      COUNT    -- Number of entries written into the file.
//      MAXJIS   -- Highest JIS code included in the file.
//
//                  KIFLAG_PINYIN (0x0001) -- Has pin yin data.
//                  KIFLAG_KOREAN (0x0002) -- Has Korean data.
//                  KIFLAG_NANORI (0x0004) -- Has nanori data.
//                  KIFLAG_EXTRA  (0x0008) -- Has extended data filed
//                  KIFLAG_EVAR   (0x0010) -- Has extended variable data
//                  KIFLAG_XREF   (0x0020) -- Has cross refference data
//
//      FIXED    -- A fixed file part.  This contains a sequency of 
//                  6355 kinfo structures representing each kanji.
//                  The structual fields are:
//                    
//                  offset      -- Points the VARIABLE part for this kanij.
//                  imi,nan,    -- Indicate the number entries of each
//                    on,kun,      type contained in the variable part.
//                    pinyin,      As data is removed to make th efile 
//                    korean,      smaller, these values are reduced (or
//                    extra        set to zero) and the correspoding 
//                                 entries are removed from the file.
//                                          
//      VARIABLE -- The variable part consists of a number of null 
//                  termianted strings writtent sequecially into the 
//                  file.  The above counters indicate the number 
//                  of each type of string.  The order of the strings
//                  is as follows:
//
//                  Korean   -- Korean data if pressent.
//                  PinYin   -- Ascii with extended charactrs to get
//                              accents.
//                  meanings -- Ascii.
//                  on-yomi  -- Compressed katakana.  The first byte is 
//                              removed, to save space.  The special
//                              code 0x1f indicates a katakana -.
//                  kun-yomi -- Compressed hiragana.  The first byte
//                              is removed, to save space.  The special
//                              code 0x1f indicates a katakana -, and 
//                              the special code 0x20 indicates the 
//                              beginning of the okurigana.
//                  nanori   -- Compressed hiragana.  Special cdoes 
//                              above are supported but not used.
//                  Extra    -- Extended field if pressent (see below for description)
//
//      EXTRA -- The extra is a binary field.  This must be the last field, because it
//               is possible to get embbedded zeros in this field.  The layout of the 
//               field is a fixed length binary structrue followed by a sequence of 
//               one-byte codes followed by the data for that argument.  The data format 
//               varies according to the idividual code field.  Much of this data 
//               is not pressent in all entries, so this type of coding takes less space.  
//               Additionally, this allows the design to change depending on what is to 
//               be added later.  
//
//              B<short>        -- Japanese For Busy People vols I-III, published by the AJLT. (High byte 
//              D<short>        -- DeRoo entry for kanji
//              E<short>        -- "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
//              F<short>        -- Frequency count.
//              G<short>        -- Kodansha Compact Kanji Guide
//              H<short>        -- Jack Halpern in his Kanji Learners Dictionary, published by Kodansha in 1999
//              I<short>        -- Spahn and Hadamitzky Kana & Kanji
//              J<short>        -- Kanji in Context by Nishiguchi and Kono                  
//              K<short>        -- Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
//              L<short>        -- "Remembering The Kanji" by James Heisig
//              N<short>        -- P.G. O'Neill's Essential Kanji 
//              O<short>        -- "Japanese Names", by P.G. O'Neill
//              Q<short>        -- Four-corners secondary code <code><resolution is stroed in the struture>
//              S<short>        -- "A Guide To Reading and Writing Japanese" edited by Florence Sakade.
//              T<short>        -- Tuttle Kanji Cards, compiled by Alexander Kask
//
//  The organization of the extra field, will always include all entries before the cross-reference
//  entries are added.  This allows the reader to mark the frist cross-reference entry and 
//  decode from there forward.  Note the cross-reference codes are all lower case, which makes
//  them easy to identifiy.
//
//              h<short>        -- Halpern dictionary cross-refernece entries.
//              n<short>        -- Nelson dictionary cross-reference entries.
//              o<short>        -- "Japanese Names", by P.G. O'Neill cross-refernece
//              k<short>        -- JIS 0208 Cross-reference
//              j<short>        -- JIS 0212 Cross-reference
//              z<short>        -- SKIP code missclassification/cross-reference 
//                                  Coded as (<error><<13) | (<type> << 10) | (<1> << 5)) | (<2>)
//
//      DROPPED ENTRY (only one of these):
//              i<short>        -- Spahn and Hadamitzky Kana & Kanji cross-reference 
//                                 coded as (rstorke << 11) | (ostroke<<6)
//

typedef unsigned short KANJI;   // Conveient definitions.
typedef unsigned char  byte;
typedef unsigned short ushort;
typedef unsigned long  ulong;

//
//  These defintions are size of buffers used to hold the string 
//  read from KANJIDIC.  There actual value has not effect on the 
//  program, and they may be changed at any time, to get rid of 
//  errors.  These values were chosen as large eough to read all
//  information, but to too large so the program ran faster.
//
#define SIZE_IMI        70      // Max size of a meaning (imi)
#define SIZE_KUN        20      // Max size of a kun-yomi
#define SIZE_NAN        10      // Max size of a nanori 
#define SIZE_ON         20      // Max size of an on-yomi
#define SIZE_PINYIN     50      // Max size for all PinYin combined.
#define SIZE_KOREAN     50      // Max size for all Korean combined.

//
//  These values actually represent the number of each type of reading
//  that thestrucrue will hold.  If you must change these values, the 
//  structure must change, so this is a real problem.
//
#define NUMBER_ON       8       // Number of on-yomi (6 is maximum used)
#define NUMBER_IMI      16      // Number of meansings (14 is maximum actually used).
#define NUMBER_NAN      32      // Number of nanori readings (25 is maximum actually used).
#define NUMBER_KUN      32      // Number of knu-yomi (16 is the maximum actually used).
#define NUMBER_STROKES  4       // Number of possilbe miscounts on strokes.
#define NUMBER_JISXREF  6       // Maximum number of JIS cross-references
#define NUMBER_SKIPXREF 6       // Number of skip missclassification codes

//
//  Coding flags for the extended field
//
#define EXTRA_FREQ      0x0001  // Frequency count.
#define EXTRA_SHKANA    0x0002  // Spahn and Hadamitzky Kana & Kanji
#define EXTRA_SHDICT    0x0004  // Spahn and Hadamitzky kanji dictionary
#define EXTRA_HENSHALL  0x0008  // "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
#define EXTRA_GAKKEN    0x0010  // Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
#define EXTRA_HEISIG    0x0020  // Remembering The Kanji" by James Heisig
#define EXTRA_ONEILL    0x0040  // "Japanese Names", by P.G. O'Neill
#define EXTRA_4C        0x0080  // Four-corners primary code <code><resolution>
#define EXTRA_4C2       0x0100  // Four-corners secondary code <code><resolution>
#define EXTRA_MDlong    0x0200  // Morohashi Daikanwajiten, long code
#define EXTRA_MDshort   0x0400  // Morohashi Daikanwajiten, short code <volume><index>

//
//  SKIP Miss-clasification / cross-reference codes:
//
#define SKIP_POSITION   1       // Position error.
#define SKIP_STROKE     2       // Stroke error.
#define SKIP_BOTH       3       // Both position and stroke error.
#define SKIP_BREEN      4       // Disagreement between Jim Breen and Jack Halpern about the code.

//
//  magic value written into the beginnin of the file.
//
#define KINFO_MAGIC     0x34A5B4D4  // This value is used to confirm that 
                                    //   this is the correct file.

//
//  Flags for the flag field.
//
#define KIFLAG_PINYIN   0x0001  // Has pin yin data.
#define KIFLAG_KOREAN   0x0002  // Has Korean data.
#define KIFLAG_NANORI   0x0004  // Has nanori data.
#define KIFLAG_EXTRA    0x0008  // Has extended data filed
#define KIFLAG_EVAR     0x0010  // Has extended variable data
#define KIFLAG_XREF     0x0020  // Has cross refference data

//
//  Encoding flags
//
#define ENCODE_EUC      1       // EUC encoding.
#define ENCODE_UTF8     2       // UTF-8 encoding
#define ENCODE_UNICODE  3       // Unicode encoding

//
//  Offset of first entry in the file.
//
#define KANJIINFO_OFFSET    (2*sizeof(ulong)+2*sizeof(ushort))

//
//  Fixed structure for data file.
//
//  This is the structure representing the fixed part of the data file
//  that JWPce is expecting.
//
#pragma pack(push,2)    // Structure to be packed at 2 byte boundaries.
                        //   this results is a smaller data size.

struct kinfo {
  ushort bushu      :8;     // Radical number (from Nelson).
  ushort strokes    :5;     // Stroke count.
  ushort on         :3;     // Number of on-yomi

  ushort grade      :4;     // Grade that Japanese learn the kanji
  ushort imi        :4;     // Number of meanings
  ushort skip_t     :3;     // Skip code: type
  ushort skip_1     :5;     //            first number

  ushort skip_2     :5;     //            second number
  ushort pinyin     :1;     // Number of PinYin lines (1 or zero).
  ushort kun        :5;     // Number of kun-yomi
  ushort nan        :5;     // Number of nanori.

  ushort extra      :1;     // Entry has extended data.
  ushort halpern    :15;    // Halpern refference number

  ushort korean     :1;     // Number of Korean lines (1 or zero)
  ushort nelson     :15;    // Nelson reference number

  ushort haig;              // Index into New Nelson dicitonary (edited by Haig)

  ulong  classical  :8;     // Classical bushu (if different from Nelson's)
  ulong  offset     :24;    // File offset to variable part of the data.
};

struct extend {
  ushort md_long;           // Morohashi Daikanwajiten long entry

  ulong  md_short1  :4;     // Morohashi Daikanwajiten short (volume)
  ulong  md_short2  :13;    // Morohashi Daikanwajiten short (kanji)
  ulong  sh_rstroke :5;     // Spahn & Hadamitzky Dictionary radical stroke count
  ulong  sh_radical :5;     // Spahn & Hadamitzky Dictionary radical (a latter)
  ulong  sh_ostroke :5;     // Spahn & Hadamitzky Dictionary other stroke count

  ulong  sh_index   :6;     // Spahn & Hadamitzky Dictionary index of kanji with radical and stroke count
  ulong  fc_main    :14;    // Four corners main entry
  ulong  fc_index   :4;     // Four corners extra entry.
  ulong  fc_index2  :4;     // Four corners extra entry (for second code)
  ulong  md_p       :1;     // Morohashi Daikanwajiten (original code 'P')
  ulong  md_x       :1;     // Morohashi Daikanwajiten (close code 'X')
};

#pragma pack(pop)

typedef kinfo  kinfo;
typedef extend extend;

//
//  This class represnets a kanji entry in the database.
//
class kinfo_big {
public:
  int parse  (char *data);      // Parse the KANJIDICT entry.
  int size   (void);            // Bytes requried to store variable part of this kanji.
  void write (FILE *file,int extended); // Write this entry into data file.
  void error (char *text);      // Generates an error message for this kanji.
  KANJI jis;                    // JIS code for character            
  short bushu;                  // Radical (Nealson).
  short classical;              // Classical (not Nelson bushu).
  short deroo;                  // DeRoo entry for kanji
  short freq;                   // Frequency count (1..~2200, 0 indicates not sent.  1 is most common
                                //                  kanji, higher numbers are less and less frequent.  
  short grade;                  // Traditional grade Japanese learn this kanji (1..6) is grade.  
                                //   8 indicates general use character.  9 Indicates for use in names 
                                //   (jinmeiyou).
  short halpern;                // Halpern dictionary number
  short nelson;                 // Nelson dictionary number
  short haig;                   // New Nelsons dictionary number
  short skip_type;              // SKIP character type (1-4)
  short skip_1;                 // SKIP first part stroke count
  short skip_2;                 // SKIP second part stroke count
  short stroke;                 // Actual stroke count.
  byte  strokes[NUMBER_STROKES];// Miscount on strokes.
  short unicode;                // Unicode value.
  short on_count;               // Number of on-yomi
  short kun_count;              // Number of kun-yomi
  short imi_count;              // Number of meannigs
  short nan_count;              // Number of nanori 
  short sh_kanakanji;           // Index into Spahn & Hadamitzky Kanji & Kana
  short sh_rstroke;             // Spahn & Hadamitzky Dictionary radical stroke count
  short sh_radical;             // Spahn & Hadamitzky Dictionary radical (a latter)
  short sh_ostroke;             // Spahn & Hadamitzky Dictionary other stroke count
  short sh_index;               // Spahn & Hadamitzky Dictionary index of kanji with radical and stroke count
  short fc_main;                // Four Corners code (main code);
  short fc_index;               // Four Corners code (index code (resolution));
  short fc2_main;               // Four Corners code (main code);
  short fc2_index;              // Four Corners code (index code (resolution));
  long  md_long;                // Morohashi Daikanwajiten long entry
  short md_short1,md_short2;    // Morohashi Daikanwajiten short entry
  short md_mark;                // Morohashi Daikanwajiten mark X, P or Blank
  short henshall;               // "A Guide To Remembering Japanese Characters" by Kenneth G. Henshall
  short gakken;                 // Gakken Kanji Dictionary ("A  New Dictionary  of Kanji Usage")
  short heisig;                 // "Remembering The Kanji" by James Heisig
  short oneill;                 // "Japanese Names", by P.G. O'Neill
  short halpern_kld;            // Jack Halpern in his Kanji Learners Dictionary, published by Kodansha in 1999
  short oneill_ek;              // P.G. O'Neill's Essential Kanji 
  short kanji_context;          // Kanji in Context by Nishiguchi and Kono
  short busy_vol;               // Japanese For Busy People vols I-III, published by the AJLT. The codes are the volume.chapter.
  short busy_chapter;           // 
  short compact_guide;          // Kodansha Compact Kanji Guide.
  short easy_way;               // The Kanji Way to Japanese Language Power by Dale Crowley
  short tuttle_card;            // Tuttle Kanji Cards, compiled by Alexander Kask
  short read_write;             // A Guide To Reading and Writing Japanese edited by Florence Sakade.
  short x_deroo;                // DeRoo cross reference code.
  short x_deroo2;   
  short x_nelson;               // Nelson cross-reference numbers
  short x_nelson2;              
  short x_oneill;               // O'Neill cross reference numbers
  short x_bushu;                // Bushu cross-reference number
  short x_halpern;              // Halpern cross-reference numbers
  short x_halpern2;             
  short x_shrstroke;            // Spahn & Hadamitzky Dicitonary cross-reference 
  short x_shostroke;
  short x_shradical;
  short x_shindex;
  short x_jis0208[NUMBER_JISXREF];  // JIS 0208 character corss-reference
  short x_jis0208c;
  short x_jis0212[NUMBER_JISXREF];  // JIS 0212 character cross-reference
  short x_jis0212c;
  short z_skipt[NUMBER_SKIPXREF];   // Skip misscode/cross reference.
  short z_skipe[NUMBER_SKIPXREF];
  short z_skip1[NUMBER_SKIPXREF];
  short z_skip2[NUMBER_SKIPXREF];
  short z_skipc;
  char pinyin[SIZE_PINYIN];         // Combind PinYin string.
  char korean[SIZE_KOREAN];         // Combind Korean string.
  byte kun[NUMBER_KUN][SIZE_KUN];   // The kun-yomi
  byte on [NUMBER_ON ][SIZE_ON ];   // The on-yomi
  byte nan[NUMBER_NAN][SIZE_NAN];   // The nanori
  char imi[NUMBER_IMI][SIZE_IMI];   // The meanings
private:
};

typedef class kinfo_big kinfo_big;








